"""Constants used by Mutagen."""

GENRES = [
    u"Blues",
    u"Classic Rock",
    u"Country",
    u"Dance",
    u"Disco",
    u"Funk",
    u"Grunge",
    u"Hip-Hop",
    u"Jazz",
    u"Metal",
    u"New Age",
    u"Oldies",
    u"Other",
    u"Pop",
    u"R&B",
    u"Rap",
    u"Reggae",
    u"Rock",
    u"Techno",
    u"Industrial",
    u"Alternative",
    u"Ska",
    u"Death Metal",
    u"Pranks",
    u"Soundtrack",
    u"Euro-Techno",
    u"Ambient",
    u"Trip-Hop",
    u"Vocal",
    u"Jazz+Funk",
    u"Fusion",
    u"Trance",
    u"Classical",
    u"Instrumental",
    u"Acid",
    u"House",
    u"Game",
    u"Sound Clip",
    u"Gospel",
    u"Noise",
    u"Alt. Rock",
    u"Bass",
    u"Soul",
    u"Punk",
    u"Space",
    u"Meditative",
    u"Instrumental Pop",
    u"Instrumental Rock",
    u"Ethnic",
    u"Gothic",
    u"Darkwave",
    u"Techno-Industrial",
    u"Electronic",
    u"Pop-Folk",
    u"Eurodance",
    u"Dream",
    u"Southern Rock",
    u"Comedy",
    u"Cult",
    u"Gangsta Rap",
    u"Top 40",
    u"Christian Rap",
    u"Pop/Funk",
    u"Jungle",
    u"Native American",
    u"Cabaret",
    u"New Wave",
    u"Psychedelic",
    u"Rave",
    u"Showtunes",
    u"Trailer",
    u"Lo-Fi",
    u"Tribal",
    u"Acid Punk",
    u"Acid Jazz",
    u"Polka",
    u"Retro",
    u"Musical",
    u"Rock & Roll",
    u"Hard Rock",
    u"Folk",
    u"Folk-Rock",
    u"National Folk",
    u"Swing",
    u"Fast-Fusion",
    u"Bebop",
    u"Latin",
    u"Revival",
    u"Celtic",
    u"Bluegrass",
    u"Avantgarde",
    u"Gothic Rock",
    u"Progressive Rock",
    u"Psychedelic Rock",
    u"Symphonic Rock",
    u"Slow Rock",
    u"Big Band",
    u"Chorus",
    u"Easy Listening",
    u"Acoustic",
    u"Humour",
    u"Speech",
    u"Chanson",
    u"Opera",
    u"Chamber Music",
    u"Sonata",
    u"Symphony",
    u"Booty Bass",
    u"Primus",
    u"Porn Groove",
    u"Satire",
    u"Slow Jam",
    u"Club",
    u"Tango",
    u"Samba",
    u"Folklore",
    u"Ballad",
    u"Power Ballad",
    u"Rhythmic Soul",
    u"Freestyle",
    u"Duet",
    u"Punk Rock",
    u"Drum Solo",
    u"A Cappella",
    u"Euro-House",
    u"Dance Hall",
    u"Goa",
    u"Drum & Bass",
    u"Club-House",
    u"Hardcore",
    u"Terror",
    u"Indie",
    u"BritPop",
    u"Afro-Punk",
    u"Polsk Punk",
    u"Beat",
    u"Christian Gangsta Rap",
    u"Heavy Metal",
    u"Black Metal",
    u"Crossover",
    u"Contemporary Christian",
    u"Christian Rock",
    u"Merengue",
    u"Salsa",
    u"Thrash Metal",
    u"Anime",
    u"JPop",
    u"Synthpop",
    u"Abstract",
    u"Art Rock",
    u"Baroque",
    u"Bhangra",
    u"Big Beat",
    u"Breakbeat",
    u"Chillout",
    u"Downtempo",
    u"Dub",
    u"EBM",
    u"Eclectic",
    u"Electro",
    u"Electroclash",
    u"Emo",
    u"Experimental",
    u"Garage",
    u"Global",
    u"IDM",
    u"Illbient",
    u"Industro-Goth",
    u"Jam Band",
    u"Krautrock",
    u"Leftfield",
    u"Lounge",
    u"Math Rock",
    u"New Romantic",
    u"Nu-Breakz",
    u"Post-Punk",
    u"Post-Rock",
    u"Psytrance",
    u"Shoegaze",
    u"Space Rock",
    u"Trop Rock",
    u"World Music",
    u"Neoclassical",
    u"Audiobook",
    u"Audio Theatre",
    u"Neue Deutsche Welle",
    u"Podcast",
    u"Indie Rock",
    u"G-Funk",
    u"Dubstep",
    u"Garage Rock",
    u"Psybient",
]
"""The ID3v1 genre list."""
